from datetime import datetime
from typing import Sequence, Callable, Any

from airflow.models import BaseOperator, SkipMixin
from airflow.providers.apache.kafka.hooks.consume import KafkaConsumerHook
from airflow.utils.context import Context


class CustomConsumeFromTopicOperator(BaseOperator, SkipMixin):
    """

    """

    def __init__(
            self,
            topics: str | Sequence[str],
            kafka_config_id: str,
            apply_function_batch: Callable,
            max_messages: int = 100,
            poll_timeout: float = 60,
            batch_timeout: float = 86400,
            **kwargs,
    ):
        super().__init__(**kwargs)

        self.topics = topics
        self.kafka_config_id = kafka_config_id
        self.apply_function_batch = apply_function_batch
        self.max_messages = max_messages
        self.poll_timeout = poll_timeout
        self.batch_timeout = batch_timeout

    def execute(self, context: Context) -> Any:
        consumer = KafkaConsumerHook(
            topics=self.topics, kafka_config_id=self.kafka_config_id
        ).get_consumer()
        messages = []
        elapsed = 0

        while elapsed < self.batch_timeout:
            start_time = datetime.now()
            timeout = min(self.poll_timeout, self.batch_timeout - elapsed)

            messages = consumer.consume(
                num_messages=self.max_messages,
                timeout=timeout,
            )

            if len(messages) > 0:
                print(f"Received {len(messages)} messages")
                break

            elapsed += (datetime.now() - start_time).total_seconds()

        if not messages:
            self.log.info("No messages. Skipping downstream tasks...")
            if downstream_tasks := context['task'].get_flat_relatives(
                upstream=False
            ):
                self.skip(
                    context['dag_run'],
                    context['ti'].execution_date,
                    downstream_tasks,
                )
            return

        result = self.apply_function_batch(messages, context)

        if result is not None:
            self.log.info(f"Result: {result}")
            context.get('ti').xcom_push(key='return_value', value=result)

        consumer.commit()
        consumer.close()
        return result


class ConsumeMessageDataOperator(CustomConsumeFromTopicOperator):
    MESSAGES_KEY = 'messages'
    MESSAGES_COUNT_KEY = 'messages_count'

    def __init__(self, **kwargs):
        args = {
            'apply_function_batch': self.messages_process,
        }
        args.update(kwargs)
        super().__init__(**args)

    @classmethod
    def messages_process(cls, messages: Sequence[Any], context) -> None:
        values = [message.value().decode('utf-8') for message in messages]
        context.get('ti').xcom_push(key=cls.MESSAGES_KEY, value=values)
        context.get('ti').xcom_push(key=cls.MESSAGES_COUNT_KEY, value=str(len(values)))



import json
from typing import List

import redis
from airflow.providers.redis.hooks.redis import RedisHook


class RedisBase:
    def __init__(self, redis_conn_id: str, redis_collection_name: str):
        self._redis_hook = RedisHook(redis_conn_id)
        self.redis_collection_name = redis_collection_name
        self.redis_conn = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self):
        if not self.redis_conn:
            self.redis_conn = self._redis_hook.get_conn()

    def close(self):
        if self.redis_conn:
            self.redis_conn.close()
            self.redis_conn = None


class OgrnRedisChecker(RedisBase):
    def is_ogrn_exists(self, ogrn: int) -> bool:
        """Проверяет наличие ОГРН в Redis множестве"""
        if not self.redis_conn:
            self.connect()

        try:
            return self.redis_conn.sismember(self.redis_collection_name, ogrn)
        except redis.ConnectionError:
            self.connect()  # Переподключение при ошибке
            return self.redis_conn.sismember(self.redis_collection_name, ogrn)


class CrossingInnOgrnRedisChecker(RedisBase):
    def get_ogrns_by_inn(self, inn: str) -> List[int]:
        if not self.redis_conn:
            self.connect()

        ogrns_str = self.redis_conn.hget(self.redis_collection_name, inn)
        ogrns = json.loads(ogrns_str) if ogrns_str else list()
        return ogrns



import inspect
import os
import sys


def get_importer_path():
    """
    Получает путь к файлу, который импортировал текущий модуль.
    Используется для определения корневого каталога проекта,
    чтобы даги могли импортировать модули из других каталогов.
    """
    stack = inspect.stack()
    for frame_info in stack[1:]:  # Пропускаем текущий фрейм
        frame = frame_info.frame
        module_name = frame.f_globals.get('__name__', '')
        file_path = frame.f_globals.get('__file__', '')

        # Пропускаем служебные модули Python
        if (
                not file_path  # Нет файла (например, REPL или exec)
                or module_name.startswith('importlib')
                or 'frozen importlib' in file_path
                or file_path.startswith('<')  # Например, <stdin>
                or file_path.startswith(os.path.dirname(os.path.realpath(__file__)))
        ):
            continue

        # Нормализуем путь
        abs_path = os.path.abspath(file_path)
        return os.path.dirname(abs_path)

    return None  # Не удалось определить


def ensure_path():
    # Получаем путь
    importer_dir = get_importer_path()

    # Добавляем в sys.path (если найден и ещё не добавлен)
    if importer_dir and importer_dir not in sys.path:
        sys.path.insert(0, importer_dir)

    print('ensure_path', importer_dir)



import datetime
import sqlalchemy


class OrmConverter:

    @classmethod
    def convert_tree_to_orm(cls, tree: dict, root_type: type):
        for key, value in list(tree.items()):
            value_type = type(value)
            data_type = cls.get_nested_type(root_type, key)

            if data_type is None:
                del tree[key]
                continue

            if value is not None:
                if isinstance(data_type, sqlalchemy.Date):
                    tree[key] = datetime.date.fromisoformat(value)
                elif isinstance(data_type, sqlalchemy.DateTime):
                    if value.endswith("Z"):
                        value = value[:-1] + "+00:00"
                    tree[key] = datetime.datetime.fromisoformat(value)

                elif isinstance(data_type, sqlalchemy.types.Enum):
                    enum_class = data_type.enum_class  # type: ignore
                    try:
                        tree[key] = enum_class[  # type: ignore
                            value
                        ]  # Преобразуем строку в Enum
                    except KeyError:
                        raise ValueError(
                            f"Invalid value '{value}' for enum {enum_class}"
                        )
                elif isinstance(data_type, sqlalchemy.Boolean):
                    tree[key] = bool(value)

            if value_type is dict:
                tree[key] = cls.convert_tree_to_orm(value, data_type)
            elif value_type is list:
                tree[key] = [cls.convert_tree_to_orm(item, data_type) for item in value]

        return root_type(**tree)

    @classmethod
    def get_nested_type(cls, parent_type, field_name) -> type | None:
        if not hasattr(parent_type, field_name):
            return None

        prop = getattr(parent_type, field_name).prop
        if hasattr(prop, "mapper"):
            return prop.mapper.class_

        return prop.expression.type
